// File: w32InetConnection.cpp 
//     Defines the w32InetConnection class for Win32 console use.
/* 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the WCTP Toolkit
 * 
 * The Initial Developer of the Original Code is WorldCom, Inc.
 * Portions created by WorldCom, Inc. are Copyright  2001
 * WorldCom, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * WorldCom, Inc. <skytel-strategic-apps@wcom.com>  
 * SEA, Inc. <jkarasz@seabase.com>
 * 
 */
/////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <errno.h>
#include <iostream.h>

// #include "stdafx.h"
#include "w32InetConnection.hpp"

//----------------------------------------------------------------------------
// w32InetConnection
//----------------------------------------------------------------------------
w32InetConnection::w32InetConnection()
{
    m_sock_fd           = INVALID_SOCKET;
    const long int usec = 1000000L; 
    m_usec_poll         = usec/10;
    m_session_limit     = 5*60; /* seconds */
}

//----------------------------------------------------------------------------
// ~w32InetConnection
//----------------------------------------------------------------------------
w32InetConnection::~w32InetConnection()
{
    fast_close();
}

//----------------------------------------------------------------------------
// is_closed
//----------------------------------------------------------------------------

bool
w32InetConnection::is_closed(void)
{
    if(m_sock_fd == INVALID_SOCKET)
    {
        return true;
    }
    return false;
}

//----------------------------------------------------------------------------
// open_conn
//----------------------------------------------------------------------------
int
w32InetConnection::open_conn(XString &error, 
						  const XString& hostname, 
						  int port)
{
	WSADATA wsaData;
    int     error_code = INET_CONN_OKAY;


	if (WSAStartup (MAKEWORD(1,1), &wsaData) != 0) 
	{
		error = "WSAStartup failed!";
		return INET_CONN_ERROR;
	}

    //
    // Clear server_addr
    //
    memset(&m_server_addr, 0, sizeof(struct sockaddr_in));

    int retVal = lookup_ipaddr(error, hostname, &m_server_addr);
	if (retVal != INET_CONN_OKAY)
	{   // failed to resolve host name
		return retVal;
	}

	// else it's okay, we resolved the host name, push on
    //
    // Set the Port
    //
    m_server_addr.sin_family = AF_INET;
    m_server_addr.sin_port = htons(port);

    //
    // Create a socket
    //
    m_sock_fd = socket(AF_INET, SOCK_STREAM, 0);
    if(m_sock_fd == INVALID_SOCKET)
    {
        int errnum = WSAGetLastError();
        error = translate_error("socket",errnum);
        m_sock_fd = INVALID_SOCKET;
        return INET_CONN_ERROR;
    }

    //
    // Recored session start
    //
    m_session_start = time(NULL);

    //
    // Connect socket to the server
    //
    int res = connect(m_sock_fd, 
        (struct sockaddr *)&m_server_addr, sizeof(m_server_addr));
    if(res != SOCKET_ERROR)
    {
        int errnum = WSAGetLastError();
        error = translate_error("socket",errnum);

        // connected immediately
        return INET_CONN_OKAY;
    }

    int errnum = WSAGetLastError();
    if(errnum != WSAEWOULDBLOCK)
    {
        error = translate_error("connect",errnum);
        return INET_CONN_ERROR;
    }

    //
    // Wait for connection to server to be established
    //
    error_code = ready_for_io(error,true,true,true);
    if(error_code != INET_CONN_OKAY)
    {
        int errnum = WSAGetLastError();
        error = translate_error("socket",errnum);
        return error_code;
    }

    //
    // Check if connection was established okay
    //
    errnum = 0;
    int errlen = sizeof(errnum);
    res = getsockopt(m_sock_fd, SOL_SOCKET, SO_ERROR, (char*)&errnum, &errlen);
    if(res == SOCKET_ERROR)
    {
        errnum = WSAGetLastError();
        error = translate_error("getsockopt",errnum);
        fast_close();
        return INET_CONN_ERROR;
    }
    else if(errnum != 0)
    {
        error = translate_error("async connect", errnum);        
        fast_close();
        return INET_CONN_ERROR;
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// close_conn
//----------------------------------------------------------------------------
int
w32InetConnection::close_conn(XString& error)
{
    //
    // Shutdown the socket connection
    //
    int res = shutdown(m_sock_fd, 2);
    if(res == SOCKET_ERROR)
    {
        int errnum = errno;
        error = translate_error("shutdown", errnum);        
        fast_close();
        WSACleanup();
        return INET_CONN_ERROR;
    }

    //
    // Wait for the buffers to flush
    //
    int error_code = ready_for_read(error);
    if(error_code != INET_CONN_OKAY)
    {
        WSACleanup();
        return error_code;
    }

    //
    // Now close the socket
    //
    res = closesocket(m_sock_fd);
    m_sock_fd = INVALID_SOCKET;
    if(res == SOCKET_ERROR)
    {
        int errnum = errno;
        error = translate_error("closesocket",errnum);        
        WSACleanup();
        return INET_CONN_ERROR;
    }

    WSACleanup();
    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// fast_close
//----------------------------------------------------------------------------
void
w32InetConnection::fast_close(void)
{
    if(m_sock_fd == INVALID_SOCKET)
    {
        return;
    }

    XString error;
    int res = closesocket(m_sock_fd);
    m_sock_fd = INVALID_SOCKET;
    if(res == SOCKET_ERROR)
    {
        int errnum = errno;
        error = translate_error("closesocket",errnum);        
    }
    return;
}

//----------------------------------------------------------------------------
// ready_for_io
//----------------------------------------------------------------------------
int
w32InetConnection::ready_for_io(
    XString& error, 
    bool read_flag, 
    bool write_flag,
	bool exp_flag)
{
    int count = 0;
    bool done = false;
    while(!done)
    {
        if(m_session_limit > 0) 
        {
            time_t current_time = time(NULL);
            time_t session_time = current_time - m_session_start;
            if(session_time > m_session_limit)
            {
                error = "session timeout";
                fast_close();
                return INET_CONN_TIMEOUT;
            }
        }

        const long int usec = 1000000L;
        struct timeval delay;
        delay.tv_sec = m_usec_poll / usec;
        delay.tv_usec = m_usec_poll % usec;

        fd_set read_fds, write_fds, exp_fds;
        fd_set *p_read_fds, *p_write_fds, *p_exp_fds;

        p_read_fds = NULL;
        if(read_flag)
        { 
            p_read_fds = &read_fds;
            FD_ZERO(p_read_fds);
            FD_SET(m_sock_fd, p_read_fds);
        }
 
        p_write_fds = NULL;
        if(write_flag)
        {
            p_write_fds = &write_fds;
            FD_ZERO(p_write_fds);
            FD_SET(m_sock_fd, p_write_fds);
        }

		p_exp_fds = NULL;
		if(exp_flag)
		{
			p_exp_fds = &exp_fds;
			FD_ZERO(p_exp_fds);
			FD_SET(m_sock_fd, p_exp_fds);
		}

        int res = select(m_sock_fd+1, 
			p_read_fds, p_write_fds, p_exp_fds, &delay);
        if(res == SOCKET_ERROR) 
        {
            int errnum = WSAGetLastError();
            // Is safe to ignore WSAEINTR here?
			if(errnum == WSAENOTSOCK)
			{
				error = "Error connecting to server, ";
				error += "make sure the server is running.";
			    fast_close();
				return INET_CONN_ERROR;
			}
            else if(errnum != WSAEINTR)
            {
                error = translate_error("select",errnum);
                fast_close();
                return INET_CONN_ERROR;
            }
        }
        else if(res > 0)
        {
            done = true;
        }
    }
    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// ready_for_read
//----------------------------------------------------------------------------
int
w32InetConnection::ready_for_read(
    XString& error)
{
    return ready_for_io(error, true, false, true);
}

//----------------------------------------------------------------------------
// ready_for_write
//----------------------------------------------------------------------------
int
w32InetConnection::ready_for_write(
    XString& error)
{
    return ready_for_io(error, false, true, true);
}

//----------------------------------------------------------------------------
// writen
//----------------------------------------------------------------------------
int
w32InetConnection::writen(
    XString &error,
    const char *data,
    int length)
{
    int count = 0;
    while(count < length)
    {
        int error_code = ready_for_write(error);
        if(error_code != INET_CONN_OKAY)
        {
            return error_code;
        }

        int write_len = length - count;
        int res = send(m_sock_fd, data + count, write_len, 0);
        if(res == SOCKET_ERROR)
        {
            int errnum = WSAGetLastError();
            if(errnum != WSAEWOULDBLOCK)
            {
                error = translate_error("send",errnum);
                fast_close();
                return INET_CONN_ERROR;
            }
        }

        count += res;
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// readn
//----------------------------------------------------------------------------
int 
w32InetConnection::readn(
    XString &error,
    XString &data,
    int length)
{
    data = "";
    int count = 0;
    while(count < length) 
    {
        int error_code = ready_for_read(error);
        if(error_code != INET_CONN_OKAY)
        {
            return error_code;
        }

        int read_len = length - count;
        if(read_len > BUFFER_SIZE)
        {
            read_len = BUFFER_SIZE;
        }

        int res = recv(m_sock_fd, m_buffer, read_len, 0);
        if(res == SOCKET_ERROR)
        {
            int errnum = WSAGetLastError();
            if(errnum != WSAEWOULDBLOCK)
            {
                error = translate_error("recv",errnum);
                fast_close();
                return INET_CONN_ERROR;
            }
        }
        else if(res == 0)
        {
            error = "client closed connection";
            fast_close();
            return INET_CONN_CLOSE;
        }
        else
        {
            data.append(m_buffer, res);
            count += res;
        }
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// read_available
//----------------------------------------------------------------------------
int
w32InetConnection::read_available(XString& error, XString& data)
{
    int error_code = ready_for_read(error);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    int res = recv(m_sock_fd, m_buffer, BUFFER_SIZE, 0);
    if(res > 0)
    {
        data.append(m_buffer, res);
    }
    else if(res == 0)
    {
        error = "client closed connection";
        fast_close();
        return INET_CONN_CLOSE;
    }
    else if(res == SOCKET_ERROR)
    {
        int errnum = WSAGetLastError();
        if(errnum != WSAEWOULDBLOCK)
        {
            error = translate_error("recv",errnum);
            fast_close();
            return INET_CONN_ERROR;
        }
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// translate_error
//----------------------------------------------------------------------------

XString 
w32InetConnection::translate_error(const char *funcname, int errnum)
{
    char *label = "";


    switch(errnum)
    {
    case WSAEINTR:
        label = "Interrupted function call";
        break;
    case WSAEBADF:
        label = "Bad file descriptor";
        break;
    case WSAEACCES:
        label = "Permission denied";
        break;
    case WSAEFAULT:
        label = "Bad address";
        break;
    case WSAEINVAL:
        label = "Invalid argument";
        break;
    case WSAEMFILE:
        label = "Too many open files";
        break;
    case WSAEWOULDBLOCK:
        label = "Resource temporarily available";
        break;
    case WSAEINPROGRESS:
        label = "Operation now in progress";
        break;
    case WSAEALREADY:
        label = "Operation already in progress";
        break;
    case WSAENOTSOCK:
        label = "Socket operation on nonsocket";
        break;
    case WSAEDESTADDRREQ:
        label = "Destination address required";
        break;
    case WSAEMSGSIZE:
        label = "Message too long";
        break;
    case WSAEPROTOTYPE:
        label = "Protocol wrong type for socket";
        break;
    case WSAENOPROTOOPT:
        label = "Bad protocol option";
        break;
    case WSAEPROTONOSUPPORT:
        label = "Protocol not supported";
        break;
    case WSAESOCKTNOSUPPORT:
        label = "Socket type not supported";
        break;
    case WSAEOPNOTSUPP:
        label = "Option not supported";
        break;
    case WSAEPFNOSUPPORT:
        label = "Protocol family not supported";
        break;
    case WSAEAFNOSUPPORT:
        label = "WSAEAFNOSUPPORT";
        break;
    case WSAEADDRINUSE:
        label = "Address already in use";
        break;
    case WSAEADDRNOTAVAIL:
        label = "Cannot assign requested address";
        break;
    case WSAENETDOWN:
        label = "Network is down";
        break;
    case WSAENETUNREACH:
        label = "Network is unreachable";
        break;
    case WSAENETRESET:
        label = "Network dropped connection on reset";
        break;
    case WSAECONNABORTED:
        label = "Software caused connection abort";
        break;
    case WSAECONNRESET:
        label = "Connection reset by peer";
        break;
    case WSAENOBUFS:
        label = "No buffer space available";
        break;
    case WSAEISCONN:
        label = "Socket is already connected";
        break;
    case WSAENOTCONN:
        label = "Socket is not connected";
        break;
    case WSAESHUTDOWN:
        label = "Cannot send after socket shutdown";
        break;
    case WSAETOOMANYREFS:
        label = "Too many references; cannot splice";
        break;
    case WSAETIMEDOUT:
        label = "Connection timed out";
        break;
    case WSAECONNREFUSED:
        label = "Connection refused";
        break;
    case WSAELOOP:
        label = "Too many levels of symbolic links";
        break;
    case WSAENAMETOOLONG:
        label = "File name too long";
        break;
    case WSAEHOSTDOWN:
        label = "Host is down";
        break;
    case WSAEHOSTUNREACH:
        label = "No route to host";
        break;
    case WSAENOTEMPTY:
        label = "Directory not empty";
        break;
    case WSAEPROCLIM:
        label = "Too many processes";
        break;
    case WSAEUSERS:
        label = "Too many users";
        break;
    case WSAEDQUOT:
        label = "Disk quota exceeded";
        break;
    case WSAESTALE:
        label = "Stale NFS file handle";
        break;
    case WSAEREMOTE:
        label = "Too many levels of remote in path";
        break;
    case WSAEDISCON:
        label = "WSAEDISCON";
        break;
    case WSASYSNOTREADY:
        label = "Network subsystem is unavailable";
        break;
    case WSAVERNOTSUPPORTED:
        label = "Winsock DLL version out of range";
        break;
    case WSANOTINITIALISED:
        label = "Successful WSAStartup not yet performed";
        break;
    case WSAHOST_NOT_FOUND:
        label = "Host not found";
        break;
    case WSATRY_AGAIN:
        label = "Nonauthoritative host not found";
        break;
    case WSANO_RECOVERY:
        label = "Nonrecoverable errors: FORMERR, REFUSED, NOTIMP";
        break;
    case WSANO_DATA:
        label = "WSANO_DATA";
        break;
    default:
        break;
    }

    //
    // Build up error message
    //
    XString error;

    // name of winsock function that last failed
    if(funcname != NULL && strlen(funcname) > 0)
    {
        error += funcname;
        error += ' ';
    }

    // winsock error number
    char buf[16];
    sprintf(buf,"%d",errnum);
    error += buf;
    error += ' ';

    // winsock error label
    if (strlen(label) > 0)
    {
        error += label;
    }

    return error;
}

//----------------------------------------------------------------------------
// lookup_ipaddr
//----------------------------------------------------------------------------
int
w32InetConnection::lookup_ipaddr(
    XString& error,
    const XString& hostname,
    struct sockaddr_in *addr)
{
    struct hostent *hentp  = NULL;
    int             herrno = INET_CONN_OKAY;

    //
    // First see if we are already given the IP address
    // as an ascii dotted-decimal string
    //
	if (addr == NULL)
	{
        error = "NULL ptr addr in lookup_ipaddr()";
        return INET_CONN_ERROR;
	}

    memset(addr, 0, sizeof(struct sockaddr_in));
    addr->sin_addr.s_addr = inet_addr(hostname.c_str());
    if(addr->sin_addr.s_addr != INADDR_NONE)
    {
        return INET_CONN_OKAY;
    }

    //
    // Perform hostname to IP address lookup
    // 
    hentp = gethostbyname(hostname.c_str());
    herrno = h_errno;

    if(hentp == NULL)
    {
        error = translate_error("gethostbyname",herrno);
        return INET_CONN_ERROR;
    }

    memset(addr, 0, sizeof(struct sockaddr_in));
    memcpy(&(addr->sin_addr), hentp->h_addr, hentp->h_length);

    return INET_CONN_OKAY;
}

