/* File:  EncodeUtil.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#include "EncodeUtil.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

unsigned char EncodeUtil::base64_encode_table[64];
unsigned char EncodeUtil::base64_decode_table[256];
unsigned char EncodeUtil::uuencode_table[64];
unsigned char EncodeUtil::uudecode_table[256];

//----------------------------------------------------------------------------
// EncodeUtil::init_tables
//----------------------------------------------------------------------------

// Note this particular method is not thread safe, but this class/library
// can be used in a threaded environment if this method is called once
// at program startup.

void 
EncodeUtil::init_tables(void)
{
    static bool first_time = true;
    int i;

    if(!first_time)
    {
        return;
    }

    first_time = false;

    //
    // fill base64_encode_table
    //

    for(i = 0; i < 26; i++) 
    {
        base64_encode_table[i] = 'A' + i;
        base64_encode_table[26 + i] = 'a' + i;
    }
    for(i = 0; i < 10; i++) 
    {
        base64_encode_table[52 + i] = '0' + i;
    }
    base64_encode_table[62] = '+';
    base64_encode_table[63] = '/';

    //
    // fill base64_decode_table
    //

    for(i = 0; i < 255; i++) 
    {
        base64_decode_table[i] = 0x80;
    }
    for(i = 'A'; i <= 'Z'; i++) 
    {
        base64_decode_table[i] = 0 + (i - 'A');
    }
    for(i = 'a'; i <= 'z'; i++) 
    {
        base64_decode_table[i] = 26 + (i - 'a');
    }
    for(i = '0'; i <= '9'; i++) 
    {
        base64_decode_table[i] = 52 + (i - '0');
    }
    base64_decode_table['+'] = 62;
    base64_decode_table['/'] = 63;
    base64_decode_table['='] = 0;

    //
    // fill uuencode_table
    //

    for(i=1; i<64; i++)
    {
        uuencode_table[i] = ' ' + i;
    }
    uuencode_table[0] = '`';

    //
    // fill uudecode_table
    //

    for(i = 0; i <= 255; i++)
    {
        uudecode_table[i] = 0x80;
    }
    uudecode_table[96] = 0;
    for(i = 33; i <= 95; i++)
    {
        uudecode_table[i] = i - 32;
    }

}

//----------------------------------------------------------------------------
// EncodeUtil::base64_encode
//----------------------------------------------------------------------------

void
EncodeUtil::base64_encode(XString &dest, const XString &src)
{
    unsigned char in[3], out[4];        
    int pos,i,n,linelen;

    init_tables();
  
    dest = "";
    linelen = 0;
    for(pos=0; pos<src.length();)
    {
        for(i=0; i<3; i++)
        {
            in[i] = '\0';
        }

        for(n=0; n<3 && pos<src.length(); n++)
        {
            in[n] = src[pos];
            pos++;
        }

        out[0] = base64_encode_table[in[0] >> 2];
        out[1] = base64_encode_table[((in[0] & 3) << 4) | (in[1] >> 4)];
        out[2] = base64_encode_table[((in[1] & 0xF) << 2) | (in[2] >> 6)];
        out[3] = base64_encode_table[in[2] & 0x3F];

        if(n<3)
        {
            out[3] = '=';
            if(n<2)
            {
                out[2] = '=';
            }
        }

        if(linelen >= 60)
        {
            dest += "\n";
            linelen = 0;
        }

        for(i=0; i<4; i++)
        {
            dest += out[i];
            linelen++;
        }
    }

    dest += "\n";
}

//----------------------------------------------------------------------------
// EncodeUtil::base64_decode
//----------------------------------------------------------------------------

bool
EncodeUtil::base64_decode(XString &dest, const XString &src)
{
    unsigned char in[4], mid[4], out[3];        
    unsigned char ch, tmp;
    int pos, i, n, m, count;
    bool done_flag, okay_flag;

    init_tables();

    okay_flag = true;
    done_flag = false;
    dest = "";

    for(pos=0; !done_flag && pos<src.length();)
    {
        for(i=0; i<4; i++)
        {
            in[i] = '\0';
        }

        for(n=0; n<4 && pos < src.length();)
        {
            ch = src[pos];
            tmp = base64_decode_table[ch];
            if(!isspace(ch))
            {
                if(tmp == 0x80)
                {
                    // found illegal base64 character 
                    okay_flag = false;
                }
                else
                {
                    in[n] = ch;
                    mid[n] = tmp;
                    n++;
                }
            }
            pos++; 
        }

        if(n == 0)
        {
            done_flag = true;
        }
        else if(n != 4)
        {
            // base64 text was truncated
            done_flag = true;
            okay_flag = false;
        }
        else
        {
            out[0] = (mid[0] << 2) | (mid[1] >> 4);
            out[1] = (mid[1] << 4) | (mid[2] >> 2);
            out[2] = (mid[2] << 6) | mid[3];

            count = 0;
            for(i=0; i<4; i++)
            {
                if(in[i] == '=')
                {
                    count++;
                }
            }

            m = 0;
            if(count == 0)
            {
                m = 3;
            }
            else if(count == 1 && in[3] == '=')
            {
                done_flag = true;
                m = 2;
            }
            else if(count == 2 && in[3] == '=' && in[2] == '=')
            {
                done_flag = true;
                m = 1;
            }
            else
            {
                // illegal set of '=' pad characters
                done_flag = true;
                okay_flag = false;
            }

            for(i=0; i<m; i++)
            {
                dest += out[i];
            }
        }
    }

    for(; okay_flag && pos < src.length(); pos++)
    {
        if(!isspace(src[pos]))
        {
            okay_flag = false;
        }
    }

    return okay_flag;
}

//----------------------------------------------------------------------------
// EncodeUtil::uuencode
//----------------------------------------------------------------------------

void
EncodeUtil::uuencode(XString &dest, const XString &src)
{
    unsigned char in[3], out[4];        
    int pos,i,n,count,num;
    bool first_flag;

    init_tables();
  
    first_flag = true;
    dest = "";
    count = 0;
    pos = 0;

    while(pos<src.length())
    {
        if(first_flag || count >= 45)
        {
            if(!first_flag)
            {
                dest += "\n";
            }

            count = 0;
            first_flag = false;

            num = src.length() - pos;
            if(num > 45)
            {
                num = 45;
            }
            dest += uuencode_table[num];
        }

        for(i=0; i<3; i++)
        {
            in[i] = '\0';
        }

        for(n=0; n<3 && pos<src.length(); n++)
        {
            in[n] = src[pos];
            count++;
            pos++;
        }

        out[0] = uuencode_table[in[0] >> 2];
        out[1] = uuencode_table[((in[0] & 3) << 4) | (in[1] >> 4)];
        out[2] = uuencode_table[((in[1] & 0xF) << 2) | (in[2] >> 6)];
        out[3] = uuencode_table[in[2] & 0x3F];

        for(i=0; i<4; i++)
        {
            dest += out[i];
        }
    }

    dest += "\n";
}

//----------------------------------------------------------------------------
// EncodeUtil::uudecode_line
//----------------------------------------------------------------------------

bool
EncodeUtil::uudecode_line(
    bool okay_flag,
    XString &dest,
    const XString &src,
    int pos)
{
    unsigned char in[4], mid[4], out[3];        
    unsigned char ch, tmp;
    int i, n, m, num;

    if(pos >= src.length())
    {
        return okay_flag;
    }

    ch = src[pos];
    pos++;
    num = uudecode_table[ch];
    if(num == 0x80)
    {
        // found non-uuencode character
        okay_flag = false;
        return okay_flag;
    }

    while(num > 0 && pos < src.length() && src[pos] != '\n')
    {
        for(i=0; i<4; i++)
        {
            mid[i] = 0;
        }

        for(n=0; n<4 && pos < src.length() && src[pos] != '\n';)
        {
            ch = src[pos];
            tmp = uudecode_table[ch];
            if(tmp == 0x80)
            {
                // found non-uuencode character
                okay_flag = false;
            } 
            else
            {
                 in[n] = ch;
                 mid[n] = tmp;
                 n++;
            }
            pos++;
        }

        out[0] = (mid[0] << 2) | (mid[1] >> 4);
        out[1] = (mid[1] << 4) | (mid[2] >> 2);
        out[2] = (mid[2] << 6) | mid[3];

        m = num;
        if(m > 3)
        {
            m = 3;
        }
        num -= m;

        if((m == 3 && n < 4) ||
           (m == 2 && n < 3) ||
           (m == 1 && n < 2))
        {
            // uuencode data is too short
            okay_flag = false;
        }

        for(i=0; i<m; i++)
        {
            dest += out[i];
        }
    }

    if(num > 0)
    {
        // uuencode data is too short
        okay_flag = false;
    }

    while(okay_flag && pos < src.length() && src[pos] != '\n')
    {
        if(!isspace(src[pos]))
        {
            // found something other than whitespace
            okay_flag = false;
        }
        pos++;
    }

    return okay_flag;
}

//----------------------------------------------------------------------------
// EncodeUtil::uudecode
//----------------------------------------------------------------------------

bool
EncodeUtil::uudecode(XString &dest, const XString &src)
{
    bool okay_flag;
    int pos;

    init_tables();

    dest = "";
    okay_flag = true;
    pos = 0;

    while(pos < src.length())
    {
        while(pos < src.length() && isspace(src[pos]))
        {
            pos++;
        }

        okay_flag = uudecode_line(okay_flag, dest, src, pos);

        while(pos < src.length() && src[pos] != '\n')
        {
            pos++;
        }
        pos++;
    }

    return okay_flag;
}

//----------------------------------------------------------------------------
// hexdump
//----------------------------------------------------------------------------

void
EncodeUtil::hexdump(XString &dest, const XString &src)
{
    int pos, i;
    dest = "";
    for(pos = 0; pos < src.size(); pos+=8)
    {
        for(i=0; i<8; i++)
        {
            if(pos+i < src.size())
            {
                char ch = src[pos+i];
                char buf[16];
                sprintf(buf,"%02X ",(unsigned int)ch);
                dest += buf; 
            }
            else
            {
                dest += "   ";                
            }
        }

        for(i=0; i<8; i++)
        {
            if(pos+i < src.size())
            {
                char ch = src[pos+i];
                if(isprint(ch) && !isspace(ch))
                {
                    dest += ch;
                }
                else
                {
                    dest += '.';
                }
            }
            else
            {
                dest += " ";
            }
        }
        
        dest += "\n";
    }
}

//----------------------------------------------------------------------------
// hex_to_char
//----------------------------------------------------------------------------

char
EncodeUtil::hex_to_char(const XString& src, int pos)
{
    char buf[4];
    unsigned int tmp;

    buf[0] = src[pos+1];
    buf[1] = src[pos+2];
    buf[2] = '\0';
    sscanf(buf,"%X",&tmp);
    if(tmp < 0 || tmp > 255)
    {
        tmp = 0;
    }
    return (char)tmp;
}

//----------------------------------------------------------------------------
// url_decode
//----------------------------------------------------------------------------

void
EncodeUtil::url_decode(XString& dest, XString &src)
{
    int pos;
    bool done;

    pos = 0;
    done = false;
    while(!done && pos < src.size())
    {
        char ch = src[pos];
        if(ch == '+')
        {
            dest += ' ';
            pos++;
        }
        else if(ch == '%')
        {
            if(pos + 2 < src.size())
            {
                dest += hex_to_char(src,pos);
                pos += 3;
            }
            else
            {
                done = true;
            }
        }
        else 
        {
            dest += ch;
            pos++;
        }
    }

    src = src.right(src.length()-pos);
}

//----------------------------------------------------------------------------
// url_encode
//----------------------------------------------------------------------------

void
EncodeUtil::url_encode(XString& dest, const XString& src)
{
    char buf[16];
    int pos;
    for(pos=0; pos<src.length(); pos++)
    {
        char ch = src[pos];
        if(ch == ' ')
        {
            dest += "+";
        }
        else if(('a'<=ch && ch<='z')||
                ('A'<=ch && ch<='Z')||
                ('0'<=ch && ch<='9'))
        {
            dest += ch;
        }
        else
        {
            sprintf(buf,"%%%02X",ch);
            dest += buf;
        }
    }
}

//----------------------------------------------------------------------------
// partial_url_decode
//----------------------------------------------------------------------------

void
EncodeUtil::partial_url_decode(XString& dest, XString &src)
{
    int pos;
    bool done;

    pos = 0;
    done = false;
    while(!done && pos < src.size())
    {
        char ch = src[pos];
        if(ch == '%')
        {
            if(pos + 2 < src.size())
            {
                dest += hex_to_char(src,pos);
                pos += 3;
            }
            else
            {
                done = true;
            }
        }
        else 
        {
            dest += ch;
            pos++;
        }
    }

    src = src.right(src.length()-pos);
}

//----------------------------------------------------------------------------
// partial_url_encode
//----------------------------------------------------------------------------

void
EncodeUtil::partial_url_encode(XString& dest, const XString& src)
{
    char buf[16];
    int pos;
    for(pos=0; pos<src.length(); pos++)
    {
        char ch = src[pos];
        if(ch == '%' || ch == '+' || (!isprint(ch) && !isspace(ch)))
        {
            sprintf(buf,"%%%02X",ch);
            dest += buf;
        }
        else
	{
	    dest += ch;
	}
    }
}
