/* File:  XString.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the XString class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml\@itqlata.com>  

*/

//-----------------------------------------------------------------------------
//
// CHANGE HISTORY:
//   Jerry Karasz - 3/13/01 - Added #ifdef _WIN32 around _vsnprintf(), not
//                        available under Linux.
//
//-----------------------------------------------------------------------------

#include "XString.hpp"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
//--------------------------------------------------------------------
// void XString::adjust(int newlen, int copyflag)
//--------------------------------------------------------------------

/* 
    This function adjusts the amount of memory allocated to hold the XString. 
It increases the size of the buffer to the next highest power of two to help
reduce the number of calls to new and delete and reduce the amount of string 
copying.  

  newlen  --  the size of the new XString
copyflag  --  if true, copy the old XString into the newly allocated memory,
              defaults to false

    This function assumes that memlen has already been set to the current
number of bytes allocated, and that str points to that memory.  If memlen
is zero, then it assumes that no memory has been allocated yet, and that
str is either uninitialized or set to NULL.  

    This function also assumes that curlen is set the the current number of
significant characters stored in that buffer.  If no significant characters
are stored, ie empty string, then curlen should be set to zero before calling
this function.  This function will set curlen to newlen before returning.

    This function should be called in all constructors so that memory does get 
allocated, and the buffer becomes null terminated for ANSI C char* conversion. 
For this to work a functions calling this routine must set curlen to newlen 
after this call.

    ** WARNING ** Don't set copy flag to true if either newlen is smaller 
than curlen, or str is NULL or str has not been initilized.  

*/

void 
XString::adjust(int newlen, bool copyflag)
{
    int len, temp;

    if(newlen < 0)
    {
        newlen = 0;
    }

    // Compute the amount of memory needed "len" to hold the contents of the 
    // new XString in a "Dobule the buffer size each time it is exceeded"
    // fashion (power of 2).  Because of the trim function, we can not assume 
    // memlen is a power of two and just double it.  Instead we must calcuate
    // the next highest power of 2.  Also, this will alocate extra room for
    // a NULL terminating character for ANSI char* conversion's sake.
    // Examples, If newlen is zero, then len gets set to one, if newlen is 2 
    // then len gets set to 4, etc.

    if(newlen < 64)
    {
        len = 64;
    }
    else
    {
        len = 1;
        temp = newlen;
        while(temp != 0) 
        {
            temp >>=1 ;
            len <<= 1;
        }
    }

    // If the new amount of required memory differs from the current
    // amount of allocated memory, then realocate.

    if(memlen != len) 
    {
        // alocate a new chunk of memory of size len
        char *p = new char[len];

        // If necessary, copy the contents of the previous string 
        // into the new memory space.
        if(copyflag) 
        {
            memcpy(p,str,curlen);
        }

        // If there was previously allocated memory, then free it.
        if(memlen != 0)
        {
            delete []str;
        }
   
        // Record the new size of allocated memory.
        memlen = len;
  
        // Update the str pointer to point to the new memory
        str = p;

    }

    // Update the current length of significant stored characters in
    // the buffer.
    curlen = newlen;

    // Null terminate the buffer for ANSI C char * conversion sake
    str[newlen] = '\0';
}

//--------------------------------------------------------------------
// XString::XString()
//--------------------------------------------------------------------
// Creates an new empty XString by default.
// ex.)  XString foo;

XString::XString()
{
    memlen = 0;
    curlen = 0;
    adjust(0);
}

//--------------------------------------------------------------------
// XString::XString(int memlen)    
//--------------------------------------------------------------------
// Creates a new empty XString with a memlen sized buffer

XString::XString(int use_memlen)    
{
    memlen = 0;
    curlen = 0;
    adjust(use_memlen);
    curlen = 0;                   // set curlen back to zero after adjust
    memset(str,'\0',use_memlen);  // fill buf with zeros to be safe.
}

//--------------------------------------------------------------------
// XString::XString(const XString &x)
//--------------------------------------------------------------------
// Creates a new XString based on an XString that already exists.
//    ex.)  XString foo(&foo2); 
// or ex.)  XString foo = foo2;
// or ex.)  void func(XString foo); // for pass by value

XString::XString(const XString& x)
{
    memlen = 0;
    curlen = 0;
    adjust(x.curlen);
    memcpy(str,x.str,curlen); 
}

//--------------------------------------------------------------------
// XString::XString(const char *x)   
//--------------------------------------------------------------------
// Creates a new XString from a given null terminated C string.
// ex.)  XString foo("foobar");
// ex.)  XString foo = "foobar";

XString::XString(const char *x)   
{
    memlen = 0;
    curlen = 0;
    adjust(strlen(x));
    memcpy(str, x, curlen);
}

//--------------------------------------------------------------------
// XString::XString(const char *x, int)   
//--------------------------------------------------------------------
// Creates a new XString from a given fixed size C string.
// Takes the callers word that size string is availble
// The string may have null characters in the middle.
// ex.)  XString foo("foobar",6);

XString::XString(const char *x, int size)   
{
    memlen = 0;
    curlen = 0;
    adjust(size);
    memcpy(str, x, curlen);
}

//--------------------------------------------------------------------
// XString::XString(const char ch)
//--------------------------------------------------------------------
// Creates a new XString containing the given char.
// ex.)  XString foo = 'a';
// ex.)  XString foo('a');

XString::XString(const char ch)
{
    memlen = 0;
    curlen = 0;
    adjust(1);
    str[0] = ch;
}

//--------------------------------------------------------------------
// XString::~XString()
//--------------------------------------------------------------------
// Frees up any allocated memory

XString::~XString()
{
    if(memlen != 0) 
    {
        delete[] str;
    }
}

//--------------------------------------------------------------------
// XString& XString::operator=(const XString &x)
//--------------------------------------------------------------------
// Overwrites the contents of an XString that already exists with
// the contents of another XString.  
// ex.) xstr1 = xstr2;

XString& 
XString::operator=(const XString& x)
{
    // watch out for the special case:  xstr1 = xstr1;
    if(this != &x) 
    { 
        adjust(x.curlen);
        memcpy(str, x.str, curlen);
    }
    return *this;
}

//--------------------------------------------------------------------
// XString& XString::operator=(const char *x)
//--------------------------------------------------------------------
// Overwrites the contents of an XString that already exists with
// the contents a null terminated C string.
// ex.) xstr1 = "foobar";
// ex.) xstr1 = (char *) p;

XString& 
XString::operator=(const char *x)
{
    // Avoid the rare case where one might try this... 
    // XString foo = "foobar"; foo = (char *)foo; 

    if(str != x) 
    {
        adjust(strlen(x));
        memcpy(str, x, curlen);
    } 
    return *this; 
}

//--------------------------------------------------------------------
// XString& XString::operator=(const char ch)
//--------------------------------------------------------------------
// Overwrites the contents of an XString that already exists with a char.
// ex.) xstr1 = 'f';
// ex.) xstr1 = (char) p;

XString& 
XString::operator=(const char ch)
{
    adjust(1);
    str[0]=ch;
    return *this;
}

//--------------------------------------------------------------------
// int XString::operator[](int i)
//--------------------------------------------------------------------
// Return the character at the given position in the string

int 
XString::operator[](int i) const
{
    int ch_wide;
    if(i < curlen) 
    {
        ch_wide = (unsigned char) str[i];
    } 
    else 
    {
        ch_wide = -1;
    }
    return ch_wide;
}

//--------------------------------------------------------------------
// XString& XString::append(const char *, int)
//--------------------------------------------------------------------

XString& 
XString::append(const char *addstr, int addlen)
{
    int olen;
    olen = curlen;
    adjust(curlen + addlen, true);
    memcpy(str + olen, addstr, addlen);  
    return *this;
}

//--------------------------------------------------------------------
// XString& XString::operator+=(const XString &x)
//--------------------------------------------------------------------
// Append an XString with another XString.
// ex.) XString foo = "foo", bar = "bar"; foo += bar;
//      cout << foo; // prints "foobar"
// This will also work for the case x+=x, because the memory 
// being copied will not overlap.

XString& 
XString::operator+=(const XString& x)
{
    int olen;
    olen = curlen;
    adjust(curlen + x.curlen, true);
    memcpy(str + olen, x.str, x.curlen);  
    return *this;
}

//--------------------------------------------------------------------
// XString& XString::operator+=(const char *x)
//--------------------------------------------------------------------
// Append an XString with a num terminated char * string
// ex.) foo += "bar";
// This also works for the rare case...
//        XString foo; foo += (char *)foo; 
// and the case 
//        XString foo = ""; foo += "";

XString& 
XString::operator+=(const char *x)
{
    int  xlen, olen;
    xlen = strlen(x);
    olen = curlen;
    adjust(xlen + curlen, true);
    memcpy(str + olen, x, xlen); 
    return *this;
}

//--------------------------------------------------------------------
// XString& XString::operator+=(const char ch)
//--------------------------------------------------------------------
// Appends the XString with the given char.
// ex.) XString foo = "zor"; foo += "k";
//      cout << foo; // prints "zork"

XString& 
XString::operator+=(const char ch)
{
    adjust(curlen + 1,true);
    str[curlen - 1] = ch;
    return *this;
}

//--------------------------------------------------------------------
// int operator==(const XString& x, const XString& y)
//--------------------------------------------------------------------
// Compares if two XStrings are equal.
// ex.)  if( foo == bar) { ... }

int 
operator==(const XString& x, const XString& y)
{
    int result;
    if(x.curlen == y.curlen) 
    {
        result = (memcmp(x.str, y.str, x.curlen) == 0);
    } 
    else 
    {
        result = 0;
    }
    return result;
}

//--------------------------------------------------------------------
// int operator==(const XString& x, const char * y)
//--------------------------------------------------------------------
// Compares if an XString and char * string are equal.
// ex.)  if( foo == "bar") { ... }

int 
operator==(const XString& x, const char *y)
{
    int result;
    int curlen = strlen(y);
    if(x.curlen == curlen) 
    {
        result = (memcmp(x.str, y, curlen) == 0);
    } 
    else 
    {
        result = 0;
    }
    return result;
}

//--------------------------------------------------------------------
// int operator!=(const XString& x, const XString& y)
//--------------------------------------------------------------------
// Compares if two XStrings are unequal.
// ex.)  if( foo != bar) { ... }

int 
operator!=(const XString& x, const XString& y)
{
    int result;
    if(x.curlen == y.curlen) 
    {
        result = (memcmp(x.str, y.str, x.curlen) != 0);
    } 
    else 
    {
        result = 1;
    }
    return result;
}

//--------------------------------------------------------------------
// int operator!=(const XString& x, const char * y)
//--------------------------------------------------------------------
// Compares if an XString and char * string are unequal.
// ex.)  if( foo != "bar") { ... }

int 
operator!=(const XString& x, const char * y)
{
    int result;
    int curlen = strlen(y);
    if(x.curlen == curlen) 
    {
        result = (memcmp(x.str, y, curlen) != 0);
    } 
    else 
    {
        result = 1;
    }
    return result;
}

//--------------------------------------------------------------------
// int operator<(const XString& x, const XString& y)
//--------------------------------------------------------------------
// Compares if one XString is lexicographically less than another. 
// ex.)  if( foo < bar) { ... }

int 
operator<(const XString& x, const XString& y)
{
    int result;
    if(x.curlen < y.curlen) 
    {
        result = (memcmp(x.str, y.str, x.curlen) <= 0);
        // check boundries
        // if x = FOO, y = FOOBAR, memcmp returns  0, result = 1 
        // if x = AOO, y = FOOBAR, memcmp returns -1, result = 1
        // if x = ZOO, y = FOOBAR, memcmp returns  1, result = 0
    } 
    else 
    {
        result = (memcmp(x.str, y.str, y.curlen) < 0);
        // check boundries
        // if x = FOO,    y = FOO, memcmp returns  0, result = 0
        // if x = AOO,    y = FOO, memcmp returns -1, result = 1 
        // if x = ZOO,    y = FOO, memcmp returns  1, result = 0
        // if x = FOOBAR, y = FOO, memcmp returns  0, result = 0
        // if x = FOOBAR, y = AOO, memcmp returns  1, result = 0
        // if x = FOOBAR, y = ZOO, memcmp returns -1, result = 1
    }
    return result;
}

//--------------------------------------------------------------------
// int operator>(const XString& x, const XString& y)
//--------------------------------------------------------------------
// Compares if one XString is lexicographically greater than another. 
// ex.)  if( foo > bar) { ... }

int 
operator>(const XString& x, const XString& y)
{
    int result;
    if(x.curlen > y.curlen) 
    {
        result = (memcmp(x.str, y.str, y.curlen) >= 0);
        // check boundries
        // if x = FOOBAR, y = FOO, memcmp returns  0, result = 1
        // if x = FOOBAR, y = AOO, memcmp returns  1, result = 1
        // if x = FOOBAR, y = ZOO, memcmp returns -1, result = 0 
    } 
    else 
    {
        result = (memcmp(x.str, y.str, y.curlen) > 0);
        // check boundries
        // if x = FOO, y = FOO,    memcmp returns  0, result = 0
        // if x = AOO, y = FOO,    memcmp returns -1, result = 0
        // if x = ZOO, y = FOO,    memcmp returns  1, result = 1
        // if x = FOO, y = FOOBAR, memcmp returns  0, result = 0 
        // if x = AOO, y = FOOBAR, memcmp returns -1, result = 0
        // if x = ZOO, y = FOOBAR, memcmp returns  1, result = 1
    }
    return result;
}

//--------------------------------------------------------------------
// static int XString::icmp(const XString& x1, const XString& x2);
//--------------------------------------------------------------------
// case insensitive compare
// returns -1 if x1 < x2, 0 if x1 == x2, 1 if x1 > x2

int
XString::icmp(const XString& x1, const XString& x2)
{
    int i; 
    int res;

    i = 0; 
    while(i < x1.curlen && i < x2.curlen && 
          toupper(x1.str[i]) == toupper(x2.str[i]))
    {
        i++;
    }

    if(i < x1.curlen && i < x2.curlen) 
    { 
        if(toupper(x1.str[i]) < toupper(x2.str[i])) 
        {
            res = -1;
        } 
        else 
        {
            res = 1;
        }
    } 
    else if(i < x1.curlen) 
    {
        res = 1;
    }
    else if(i < x2.curlen) 
    {
        res = -1;
    }
    else 
    {
        res = 0;
    }    
    return res;
}

//--------------------------------------------------------------------
// char* XString::data() 
//--------------------------------------------------------------------
// Returns a const unsigned char * pointer to the memory location where 
// the contents of the XString begins.
//
// This function now returns a NULL terminated C String, however,
// if the string has a null character in the middle, then the ANSI
// char* string won't see the whole string. 
//
// It is VERY important that the XString varible does not go
// out of scope (destruct) while this pointer is being used, or else
// a Invalid Memory Access (SEGV) could occur.
//

char* 
XString::data() const
{
  return str;
}

//--------------------------------------------------------------------
// XString::operator const char*() const
//--------------------------------------------------------------------
// Returns a const char * pointer of the memory location where the
// XString begins. Same warnings apply as the data() function, see above.

char* 
XString::c_str() const
{
  return str;
}


//--------------------------------------------------------------------
// XString::operator const char*() const
//--------------------------------------------------------------------
// Returns a const char * pointer of the memory location where the
// XString begins. Same warnings apply as the data() function, see above.

XString::operator const char*() const
{
    return (const char*) str;
}

//--------------------------------------------------------------------
// int XString::length() const
//--------------------------------------------------------------------
// Returns the current length of the XString

int 
XString::length(void) const
{
    return (int) curlen; 
}

//--------------------------------------------------------------------
// int XString::size() const
//--------------------------------------------------------------------
// Returns the current length of the XString

int 
XString::size(void) const
{
    return (int) curlen; 
}

//--------------------------------------------------------------------
// bool XString::empty(void) const;         
//--------------------------------------------------------------------
// returns true if size()==0

bool 
XString::empty(void) const
{
    return (bool)(curlen == 0);
}

//--------------------------------------------------------------------
// XString XString::left(int size)
//-------------------------------------------------------------------- 

XString 
XString::left(int size)
{ 
    if(size > curlen) 
    {
        size = curlen;
    }
    return XString(str,size);
}

//--------------------------------------------------------------------
// XString XString::right(int size)
//--------------------------------------------------------------------

XString 
XString::right(int size)
{
    int start;
    if(size > curlen) 
    {
        size = curlen;
    }
    start = curlen - size;
    return XString(str+start, size);
}

//--------------------------------------------------------------------
// bool XString::is_white_space(void) const
//--------------------------------------------------------------------
// Does the string contain just white space?

bool
XString::is_white_space(void) const
{
    int i;
    for(i=0; i<curlen; i++)    
    {
        if(!isspace(str[i]))
        {
            return false;
        }
    }
    return true;
}

//--------------------------------------------------------------------
// static bool XString::is_white_space(const char *); 
//--------------------------------------------------------------------
// Does the char * string contain just white space?

bool 
XString::is_white_space(const char *pString)
{
    int i,len;
    len = strlen(pString);
    for(i=0; i<len; i++)
    {
        if(!isspace(pString[i]))
        {
            return false;
        }
    }
    return true;
}


//--------------------------------------------------------------------
// bool XString::has_white_space(void) const
//--------------------------------------------------------------------
// Does the string contain ANY white space?

bool
XString::has_white_space(void) const
{
    int i;
    for(i=0; i<curlen; i++)    
    {
        if(isspace(str[i]))
        {
            return true;
        }
    }
    return false;
}

//--------------------------------------------------------------------
// static bool XString::has_white_space(const char *); 
//--------------------------------------------------------------------
// Does the char * string contain ANY white space?

bool 
XString::has_white_space(const char *pString)
{
    int i,len;
    len = strlen(pString);
    for(i=0; i<len; i++)
    {
        if(isspace(pString[i]) )
        {
            return true;
        }
    }
    return false;
}



//--------------------------------------------------------------------
// XString XString::both_trim(void) const
//--------------------------------------------------------------------
// Return a new string with the leading and trailing white space removed

XString
XString::both_trim(void) const
{
    int start, stop;

    start = 0;
    while(start < curlen && isspace(str[start]))
    {
        start++;
    }

    stop = curlen - 1;
    while(stop >= start && isspace(str[stop]))
    {
        stop--;
    }

    return XString(str + start, stop - start + 1);
}

//--------------------------------------------------------------------
// bool XString::is_multiline(void) const
//--------------------------------------------------------------------
// Checks if string has newline or linefeed characters in it

bool
XString::is_multiline(void) const
{
    int i;
    for(i=0; i<curlen; i++)
    {
        if(str[i] == '\n' || str[i] == '\r')
        {
            return true;
        }
    }
    return false;
}
//--------------------------------------------------------------------
// const int XString::find(const char x, int begin) const;         
//--------------------------------------------------------------------
//returns positon in string where char is found

int XString::find(const char x, int begin) const
{
    int i; 

    i = begin; 
    while(i < curlen &&  toupper(str[i]) != toupper(x))
    {
        i++;
    }

    if (i > curlen)
        return -1;
    else 
        return i;
}

//--------------------------------------------------------------------
// int XString::find(const XString &x, int begin) const;         
//--------------------------------------------------------------------
//returns positon in string where x is found
int XString::find(const XString &x, int begin) const         
{
    int length = x.length();
    if (begin > length)
        return -1;


    // find first matching substring
    char *lpsz = strstr(str + begin, x);

    // return -1 for not found, distance from beginning otherwise
    return (lpsz == NULL) ? -1 : (int)(lpsz - str);
}


//--------------------------------------------------------------------
// int XString::find(const char *pFindStr, int begin) const;         
//--------------------------------------------------------------------
//returns positon in string where x is found
int XString::find(const char *pFindStr, int begin) const         
{
    if(pFindStr == NULL)
    {
        return -1;
    }

    //kah int length = strlen(pFindStr);
    if (begin > length())
    {
        return -1;
    }


    // find first matching substring
    char *lpsz = strstr(str + begin, pFindStr);

    // return -1 for not found, distance from beginning otherwise
    return (lpsz == NULL) ? -1 : (int)(lpsz - str);

}


XString &XString::format(const char *pFormatStr, ...)
{
    char *pStringBuf = new char[XSTRING_MAX_FORMAT_OUTPUT_STRING_LENGTH + 1];

    va_list valMarker;
    va_start(valMarker, pFormatStr);

    // _vsnprintf() is not available on Linux JDK 3/13/01
#ifdef _WIN32
    _vsnprintf(pStringBuf, XSTRING_MAX_FORMAT_OUTPUT_STRING_LENGTH, pFormatStr, 
        valMarker);
#else
    vsnprintf(pStringBuf, XSTRING_MAX_FORMAT_OUTPUT_STRING_LENGTH, pFormatStr, 
        valMarker);
#endif
    va_end(valMarker);         

    *(pStringBuf + XSTRING_MAX_FORMAT_OUTPUT_STRING_LENGTH) = 0;
    *this = pStringBuf;
    delete [] pStringBuf;
    return *this;
}
