/* File:  wctpXmlPayload.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#include "wctpXmlPayload.hpp"
#include <stdio.h>

//////////////////////////////////////////////////////////////////////////////
// Constructor

wctpXmlPayload::wctpXmlPayload()
{
    Init();
}

//////////////////////////////////////////////////////////////////////////////
// Copy Constructor

wctpXmlPayload::wctpXmlPayload(
    const wctpXmlPayload &rval)
{
    Copy(rval);
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

wctpXmlPayload::~wctpXmlPayload()
{
    Free();
}

//////////////////////////////////////////////////////////////////////////////
// Assignment Operator

wctpXmlPayload&
wctpXmlPayload::operator=(
    const wctpXmlPayload &rval)
{
    Free();
    Copy(rval);
    return *this;
}

//////////////////////////////////////////////////////////////////////////////
// Initialize member elements to default values

void
wctpXmlPayload::Init(void)
{
    wctpXmlOperation::Init();
    
    m_PayloadType = WCTP_XML_UNDEFINED;
    m_pMCR = NULL;
    m_pAlphanumeric = NULL;
    m_pTransparentData = NULL;

    m_title = PAYLOAD_STRING;

    m_xmlOperation = m_xmlPrefix;
    m_xmlOperation += "-";
    m_xmlOperation += m_title;


}

//////////////////////////////////////////////////////////////////////////////
// Deep copy member elements

void
wctpXmlPayload::Copy(const wctpXmlPayload &rval)
{
    wctpXmlOperation::Copy(rval);

    //
    // Copy object type
    //

    m_PayloadType = rval.m_PayloadType;

    //
    // Copy pMCR object if available (maybe holds)
    //

    if(rval.m_pMCR != NULL)
    {
        m_pMCR = new wctpXmlMCR(
            *rval.m_pMCR);
    }
    else
    {
        m_pMCR = NULL;
    }

    //
    // Copy pAlphanumeric object if available (maybe holds)
    //

    if(rval.m_pAlphanumeric != NULL)
    {
        m_pAlphanumeric = new wctpXmlAlphanumeric(
            *rval.m_pAlphanumeric);
    }
    else
    {
        m_pAlphanumeric = NULL;
    }

    //
    // Copy pTransparentData object if available (maybe holds)
    //

    if(rval.m_pTransparentData != NULL)
    {
        m_pTransparentData = new wctpXmlTransparentData(
            *rval.m_pTransparentData);
    }
    else
    {
        m_pTransparentData = NULL;
    }

    m_xmlOperation = rval.m_xmlOperation;

}


//////////////////////////////////////////////////////////////////////////////
// Free any objects being held

void
wctpXmlPayload::Free(void)
{
    wctpXmlOperation::Free();

    if(m_pMCR != NULL) 
    {
        delete m_pMCR;
        m_pMCR = NULL;
    }
    if(m_pAlphanumeric != NULL) 
    {
        delete m_pAlphanumeric;
        m_pAlphanumeric = NULL;
    }
    if(m_pTransparentData != NULL) 
    {
        delete m_pTransparentData;
        m_pTransparentData = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// Clear object (reset object as if it were just constructed)

void
wctpXmlPayload::Clear(void)
{
    Free();
    Init();
}


//////////////////////////////////////////////////////////////////////////////
// Construct XML string for this object using given context 

XString 
wctpXmlPayload::AssembleOperation(XmlContext& ctx)
{
    XString dest;
    XString temp;

    temp = "<";
    temp += m_xmlOperation;
    temp += ">";
    dest += ctx.OpenIndent(temp);

    if(m_PayloadType == WCTP_XML_MCR && 
       m_pMCR != NULL)
    {
        dest += m_pMCR->AssembleOperation(ctx);
    }
    if(m_PayloadType == WCTP_XML_ALPHANUMERIC && 
       m_pAlphanumeric != NULL)
    {
        dest += m_pAlphanumeric->AssembleOperation(ctx);
    }
    if(m_PayloadType == WCTP_XML_TRANSPARENT_DATA && 
       m_pTransparentData != NULL)
    {
        dest += m_pTransparentData->AssembleOperation(ctx);
    }

    temp = "</";
    temp += m_xmlOperation;
    temp += ">";
    dest += ctx.CloseIndent(temp);

    return dest;     
}

//////////////////////////////////////////////////////////////////////////////
// Parse switch for choice among set of objects (must match one)

wctp_parse_t 
wctpXmlPayload::ParseObjectSwitch(
    XString& msg,
    XmlNode *pTree, 
    int& pos)
{
    XString mcr_string = m_xmlPrefix;
    mcr_string += "-";
    mcr_string += MCR_STRING;

    XString alpha_string = m_xmlPrefix;
    alpha_string += "-";
    alpha_string += ALPHANUMERIC_STRING;

    XString trans_string = m_xmlPrefix;
    trans_string += "-";
    trans_string += TRANSPARENT_DATA_STRING;;


    wctp_parse_t res = WCTP_PARSE_ERROR;
    if(pos >= pTree->children.size())
    {
        msg = "An element is required inside ";
        msg += m_xmlOperation;
        return res;
    }
    XmlNode *pChild = pTree->children[pos];
    if(pChild->element == mcr_string)
    {
        HoldMCR();
        pos++;
        res = m_pMCR->ParseObject(msg,pChild);
    }
    else if(pChild->element == alpha_string)
    {
        HoldAlphanumeric();
        pos++;
        res = m_pAlphanumeric->ParseObject(msg,pChild);
    }
    else if(pChild->element == trans_string)
    {
        HoldTransparentData();
        pos++;
        res = m_pTransparentData->ParseObject(msg,pChild);
    }
    else 
    {
        msg = "unexpected element |";
        msg += pChild->element;
        msg += "| found inside ";
        msg += m_xmlOperation;
        res = WCTP_PARSE_ERROR;
    }
    return res;
}

//////////////////////////////////////////////////////////////////////////////
// Parse XML tree for this object 

wctp_parse_t 
wctpXmlPayload::ParseObject(XString& msg,XmlNode *pTree)
{
    //
    // make sure the element matches me
    //

    if(m_xmlOperation != pTree->element)
    {
        msg = "element mismatch, expected ";
        msg += m_xmlOperation;
        msg += ", but got |";
        msg += pTree->element;
        msg += "|";
        return WCTP_PARSE_ERROR;  
    }

    //
    // make sure no attributes where specified 
    //

    if(pTree->attributes.size() > 0)
    {
        msg = m_xmlOperation;
        msg += " does not have any attributes";
        return WCTP_PARSE_ERROR;
    }

     //
     // make sure no pcdata
     //

     XString pcdata = pTree->pcdata.both_trim();
     if(pcdata.size() > 0)
     {
        msg = m_xmlOperation;
        msg += " does not take pcdata";
        return WCTP_PARSE_ERROR;
     }

    //
    // Load in child objects
    //

    int pos = 0;
    wctp_parse_t res;

    //
    // Load in an object from a set
    //

    res = ParseObjectSwitch(msg, pTree, pos);
    if(res != WCTP_PARSE_OKAY)
    { 
        return res; 
    }

    //
    // Should not be any remaining child objects
    //

    if(pos < pTree->children.size())
    {
        XmlNode *pChild = pTree->children[pos];
        msg = m_xmlOperation;
        msg += " element does not accept a |";
        msg += pChild->element;
        msg += "| child element";
        return WCTP_PARSE_ERROR;
    }

    return WCTP_PARSE_OKAY;
}

//////////////////////////////////////////////////////////////////////////////
// Accessor to find which object type is contained

wctp_xml_t 
wctpXmlPayload::GetPayloadType(void)
{
    return m_PayloadType;
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for pointer to MCR object
// This function will never return NULL unless new fails...

wctpXmlMCR*
wctpXmlPayload::GetMCR(void) 
{
    
    return m_pMCR;
}

//////////////////////////////////////////////////////////////////////////////
// Specify container to hold MCR object 

void
wctpXmlPayload::HoldMCR(void) 
{
    m_PayloadType = WCTP_XML_MCR;
    if(m_pMCR == NULL)
    {
        m_pMCR = new wctpXmlMCR;
        m_pMCR->SetPrefix(m_xmlPrefix);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for pointer to Alphanumeric object
// This function will never return NULL unless new fails...

wctpXmlAlphanumeric*
wctpXmlPayload::GetAlphanumeric(void) 
{
    
    return m_pAlphanumeric;
}

//////////////////////////////////////////////////////////////////////////////
// Specify container to hold Alphanumeric object 

void
wctpXmlPayload::HoldAlphanumeric(void) 
{
    m_PayloadType = WCTP_XML_ALPHANUMERIC;
    if(m_pAlphanumeric == NULL)
    {
        m_pAlphanumeric = new wctpXmlAlphanumeric;
        m_pAlphanumeric->SetPrefix(m_xmlPrefix);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for pointer to TransparentData object
// This function will never return NULL unless new fails...

wctpXmlTransparentData*
wctpXmlPayload::GetTransparentData(void) 
{
    return m_pTransparentData;
}

//////////////////////////////////////////////////////////////////////////////
// Specify container to hold TransparentData object 

void
wctpXmlPayload::HoldTransparentData(void) 
{
    m_PayloadType = WCTP_XML_TRANSPARENT_DATA;
    if(m_pTransparentData == NULL)
    {
        m_pTransparentData = new wctpXmlTransparentData;
        m_pTransparentData->SetPrefix(m_xmlPrefix);
    }
}


	
